import base64
import botocore
import hmac
import hashlib
import json
import os
import re
import requests
import pandas as pd
from IPython.display import HTML
from IPython.display import display

from datetime import datetime, timezone, timedelta
from .enums import CheckStatusCode, HTTPStatusCode
from .cassandra_utils import PagedResultHandler
from sqlalchemy import text
from typing import Dict

class SDTcloud:
    """SDTcloud is a class to access SDT-Cloud resources.
    
    Methods:
        __init__(self)
        get_timestreamdb_table_info(self)
        get_timestreamdb_column_info(self, database_name, table_name)
        get_timestreamdb_data(self, database_name, table_name, query_condition)
        get_postgresdb_table_info(self)
        get_postgresdb_column_info(self, table_name)
        get_postgresdb_data(self, table_name, query_condition)
        init_influxdb(self)
        get_measurements_from_influx_bucket(self)
        show_fields_from_measurement(self, bucket, measurement)
        get_data_from_measurement(self, bucket, measurement, query_condition)
        git_config_init(self)
        s3_file_upload(self, local_path, bucket_name)
        s3_file_download(self, bucket_name, object_key, local_path)
        s3_list_objet(self, bucket_name)
        s3_get_object(self, bucket_name, file_name)
        s3_delete_object(self, bucket_name, file_name)
        s3_create_bucket(self, bucket_name)
        s3_delete_bucket(self, bucket_name)
        s3_bucket_list(self)
        get_projects(self)
        get_project_code_by_idx(self, idx, df)
        set_project(self, idx)
        get_user_repository(self)
        get_public_repository(self)
        git_repo_init_folder(self, path='.')
        git_repo_add_remote_url(self, remote_url, path='.')
        clone_user_repository(self, clone_url)
        add_stage_file(self, file_path_for_add_stages='.')
        restore_staged_file(self, file_path_for_restore_stages)
        commit_repository(self, commit_message)
        push_repository(self, push_file_path='.')
        pull_repository(self, repository_url)
        create_repository(self, repo_name, private_flag=True)
    """
    def __init__(self):
        """SDTcloud Class Constructor to initialize the objects.
        """
        # self.url = f"http://datalake-internal-query-service.sdt-cloud.svc.cluster.local:8080"  # datalake url
        self.url = None
        # self.giteamanager_url = f'http://gitea-manager.sdt-cloud.svc.cluster.local:8010'  # gitea-manager url
        self.giteamanager_url = None  # gitea-manager url
        self.giteamanager_url_regex = r"((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)(/)?"  # gitea url regex
        self.organizationId = ""
        self.email = ""

        # InfluxDB
        self.influxdb_client = None
        self.influx_url = ""
        self.influx_organization = ""
        self.influx_token = ""
        self.influx_bucket = ""
        self.influx_measurement = ""

        self.project_dataframe = None  # project DataFrame
        self.asset_dataframe = None  # asset DataFrame

        self.current_project_idx = None  # current project idx
        self.current_project_name = None  # current project name
        self.current_device = None  # current device

        self.gitea_user_name = None
        ############################################# AWS START ###################################################
        # Cube
        self.cubing_job_dataframe = None  # cubing 작업 dataframe
        self.cubing_history_dataframe = None  # cubing 작업이력 dataframe
        self.cube_bucket = ""
        self.cube_access_key = ""
        self.cube_secret_key = ""
        self.cube_resion = ""

        # S3
        self.s3_url = ''
        self.s3_access_key = ''
        self.s3_secret_key = ''
        self.s3_region = ''
        self.s3_bucket = ''

        # TimeStreamDB(TODO: ProjectCode로 접속정보 조회 시 조회결과로 받아야 할 것)
        self.timestreamdb_access_key = None
        self.timestreamdb_secret_access_key = None
        self.timestreamdb_region = None
        self.timestreamdb_write_client = None
        self.timestreamdb_database = None

        # PostgreSQL(TODO: ProjectCode로 접속정보 조회 시 조회결과로 받아야 할 것)
        self.postgresdb_url = None
        self.postgresdb_username = None
        self.postgresdb_password = None
        self.postgresdb_database = None
        self.postgresdb_port = '5432'
        self.postgresdb_client = None
        ############################################# AWS END ###################################################

        # env
        variables_dict = self.get_variable_from_script()
        # self.gitea_user_name = variables_dict.get('GITEA_USER_NAME').replace('"', '')  # gitea user name
        self.endpointexplorer_url = variables_dict.get('ENDPOINT_EXPLORER_URL').replace('"', '')  # endpoint explorer url
        self.organizationId = variables_dict.get('ORGANIZATION_ID').replace('"', '')  # user organization id
        self.giteamanager_url = variables_dict.get('PKG_GITEA_MANAGER_URL').replace('"', '')  # gitea manager url
        self.url = variables_dict.get('PKG_PROJECT_MANAGER_URL').replace('"', '')  # project manager url
        self.pjm_hmac_secret_key = variables_dict.get('PJM_HMAC_SECRET_KEY').replace('"', '')  # hmac secret key
        self.cube_manager_url = variables_dict.get('CUBE_MANAGER_URL').replace('"', '')
        self.cube_access_key = variables_dict.get('CUBE_ACCESS_KEY').replace('"', '')  # cube 조회용 access key
        self.cube_secret_key = variables_dict.get('CUBE_SECRET_KEY').replace('"', '')  # cube 조회용 secret key
        self.cube_region = variables_dict.get('CUBE_REGION').replace('"', '')
    
    def get_variable_from_script(self):
        variables = {}
        file_path = '/home/ec2-user/SageMaker/.env'

        try:
            with open(file_path, 'r') as script_file:
                # 스크립트 파일의 각 줄을 읽음
                for line in script_file:
                    # 정규 표현식을 사용하여 변수와 값을 추출
                    match = re.match(r'^\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*=\s*(.*)\s*$', line)
                    if match:
                        variable_name = match.group(1)
                        variable_value = match.group(2)
                        variables[variable_name] = variable_value
        except FileNotFoundError:
            print(f"파일 '{file_path}'을 찾을 수 없습니다.")
        except Exception as e:
            print(f"오류 발생: {e}")

        return variables
    
    def _generate_header(self):
        try:
            pjm_timestamp = datetime.now(timezone.utc).strftime('%Y%m%d%H%M%S')  # timestamp(utc)
            message = f'{self.organizationId}{pjm_timestamp}'  # message(organizationId + timestamp)

            # hmac 생성(sha256)
            m = hmac.new(self.pjm_hmac_secret_key.encode('utf-8'), message.encode('utf-8'), hashlib.sha256)
            digest = m.digest()

            # header
            headers = {
                "Content-Type": "application/json",
                "Organization-Id": self.organizationId,
                "Timestamp": pjm_timestamp,
                "HMAC": base64.b64encode(digest).decode('utf-8')
            }

            return headers
        except Exception as e:
            raise Exception(e)

    ###################################################### Cube Start ######################################################
    def get_cube_job_list(self):
        """ Show Cubing job list from cube-manager
        
        Returns: dataframe

        Raises:
            AttributeError: raise when timestream write client is None
            Exception: raise other exception
        """
        try:
            headers = self._generate_header()

            res = requests.request('GET', f'{self.cube_manager_url}/cube-manager/v1/storage/jobs', headers=headers)

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                self.cubing_job_dataframe = pd.DataFrame(result_data)
                return self.cubing_job_dataframe
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    
    def get_cube_history_list(self, job_idx):
        """ get cube history use job_idx

        Args:
            job_idx (_type_): job idx from job dataframe

        Raises:
            Exception: _description_

        Returns:
            _type_: _description_
        """
        try:
            headers = self._generate_header()
            job_name = self.cubing_job_dataframe.iloc[job_idx]['job_name']
            data = {
                'access_key': self.cube_access_key,
                'secret_key': self.cube_secret_key,
                'job_name': job_name,
                'region': self.cube_region,
                'bucket_name': self.organizationId
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/cube/histories', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                self.cubing_history_dataframe = pd.DataFrame(result_data, columns=['file_name', 'size(Bytes)', 'time'])
                return self.cubing_history_dataframe
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    
    def download_cube_file(self, cube_idx):
        try:
            headers = self._generate_header()
            download_key = self.cubing_history_dataframe.iloc[cube_idx]['file_name']
            data = {
                'access_key': self.cube_access_key,
                'secret_key': self.cube_secret_key,
                'download_file_name': download_key,
                'region': self.cube_region,
                'bucket_name': self.organizationId
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/cube/histories/file-download', headers=headers, data=json.dumps(data))

            download_file_name = download_key.split('/')[-1] if "/" in download_key else download_key

            if res.status_code == HTTPStatusCode.OK.value:
                with open(f'./{download_file_name}', 'wb') as f:
                    f.write(res.content)
                return {"message": f"file download in ./{download_file_name}"}

            print(res.status_code, res.json().get('message'))

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    
    def merge_cube_file(self, cube_file_list):
        try:
            headers = self._generate_header()
            data = {
                'access_key': self.cube_access_key,
                'secret_key': self.cube_secret_key,
                'cube_history_list': cube_file_list,
                'region': self.cube_region,
                'bucket_name': self.organizationId
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/cube/histories/merge-data', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                result_df = pd.DataFrame(result_data)
                return result_df
            print(res.status_code, res.json().get('message'))

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    ###################################################### Cube End ######################################################
    
    ###################################################### TimeStreamDB ######################################################
    def get_timestreamdb_table_info(self):
        """ Show table information by database into DataFrame.
        
        Returns: None

        Raises:
            AttributeError: raise when timestream write client is None
            Exception: raise other exception
        """
        try:
            headers = self._generate_header()

            data = {
                'access_key': self.timestreamdb_access_key,
                'secret_key': self.timestreamdb_secret_access_key,
                'region': self.timestreamdb_region,
                'database': self.timestreamdb_database
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/tsdb/tables', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data, columns=['database', 'table'])
                return df
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)

    def get_timestreamdb_column_info(self, table_name):
        """ Show column information of table into DataFrame.

        Args:
            table_name (str): timestream table_name

        Raises:
            AttributeError: raise when timestream write client is None
            Exception: raise other exception

        Returns: None
        """
        try:
            headers = self._generate_header()

            data = {
                'access_key': self.timestreamdb_access_key,
                'secret_key': self.timestreamdb_secret_access_key,
                'region': self.timestreamdb_region,
                'database': self.timestreamdb_database,
                'table': table_name
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/tsdb/tables/column', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data, columns=['column_name', 'column_type'])
                return df
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)

    def get_timestreamdb_data(self, table_name, query_condition):
        """ Get timestreamdb data of table use query_condition into DataFrame.

        Args:
            table_name (str): timestreamdb table_name
            query_condition (dict): query_condition

        Raises:
            AttributeError: raise when timestream query client is None
            AttributeError: raise start_datetime in query_condition format is wrong.
            AttributeError: raise end_datetime in query_condition format is wrong.
            Exception: raise other exception

        Returns:
            pd.DataFrame: query result
        """
        try:
            headers = self._generate_header()

            data = {
                'access_key': self.timestreamdb_access_key,
                'secret_key': self.timestreamdb_secret_access_key,
                'region': self.timestreamdb_region,
                'database': self.timestreamdb_database,
                'table': table_name,
                'query_condition': query_condition
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/tsdb/query', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data)
                return df
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    ###################################################### TimeStreamDB ######################################################
    ###################################################### PostgreSQL ######################################################
    def get_postgresdb_table_info(self):
        """postgresqldb get table information
        
        Raises:
            AttributeError: raise when postgresdb client is None
            Exception: raise other exception
        """
        try:
            headers = self._generate_header()

            data = {
                "host": self.postgresdb_url,
                "port": self.postgresdb_port,
                "user": self.postgresdb_username,
                "password": self.postgresdb_password,
                "database": self.postgresdb_database
            }
            
            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/pg/tables', headers=headers, data=json.dumps(data))
            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data, columns=['database', 'table'])
                return df
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    
    def get_postgresdb_column_info(self, table_name):
        """postgresqldb get column information of table.
        
        Raises:
            AttributeError: raise when postgresdb client is None
            Exception: raise other exception
        """
        try:
            headers = self._generate_header()

            data = {
                "host": self.postgresdb_url,
                "port": self.postgresdb_port,
                "user": self.postgresdb_username,
                "password": self.postgresdb_password,
                "database": self.postgresdb_database,
                "table": table_name
            }
            
            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/pg/tables/column', headers=headers, data=json.dumps(data))
            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data, columns=['column_name', 'column_type'])
                return df
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    
    def get_postgresdb_data(self, table_name, query_condition):
        """Get postgresdb data of table use query_condition into DataFrame.

        Args:
            table_name (str): postgresdb table_name
            query_condition (dict): query_condition

        Raises:
            AttributeError: raise when postgresdb client is None
            AttributeError: raise start_datetime in query_condition format is wrong.
            AttributeError: raise end_datetime in query_condition format is wrong.
            Exception: raise other exception

        Returns: 
            pd.DataFrame: query result
        """        
        try:
            headers = self._generate_header()

            data = {
                "host": self.postgresdb_url,
                "port": self.postgresdb_port,
                "user": self.postgresdb_username,
                "password": self.postgresdb_password,
                "database": self.postgresdb_database,
                "table": table_name,
                "query_condition": query_condition
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/pg/query', headers=headers, data=json.dumps(data))
            
            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data)
                return df
            elif res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return None

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    ###################################################### PostgreSQL ######################################################
    ###################################################### InfluxDB ######################################################

    def init_influxdb(self):
        """ influxdb client Init

        Raises:
            Exception: raise other exception
        """        
        try:
            import influxdb_client
            self.influxdb_client = influxdb_client.InfluxDBClient(
                url=self.influx_url,
                token=self.influx_token,
                org=self.influx_organization
            )
            print('InfluxDB init Success!!')
        except Exception as e:
            raise Exception(e)

    def get_measurements_from_influx_bucket(self):
        """ Get measurements from influx bucket.

        Raises:
            AttributeError: raise when influxdb client is None
            Exception: raise other exception

        Returns:
            pd.DataFrame : query result
        """        
        if self.influxdb_client is None:
            raise AttributeError('Please init_influxdb function run!!')

        try:
            db_result = self.influxdb_client.buckets_api().find_buckets().buckets
            database_list = [bucket.name for bucket in db_result if bucket.name not in ['_monitoring', '_tasks']]

            table_dict = {}
            for database in database_list:
                query = f'import "influxdata/influxdb/schema" schema.measurements(bucket: "{database}")'
                result = self.influxdb_client.query_api().query(query)
                measurement_list = [row.values["_value"] for table in result for row in table]
                table_dict[database] = measurement_list

            new_data = []
            for key, values in table_dict.items():
                for value in values:
                    new_data.append([key, value])

            df = pd.DataFrame(new_data, columns=['bucket', 'measurement'])
            return df
        except Exception as e:
            raise Exception(e)
    
    def show_fields_from_measurement(self, bucket, measurement):
        """ Get fields from influx measurement.

        Args:
            bucket (str): influx bucket name
            measurement (str): influx measurement name

        Raises:
            AttributeError: raise when influxdb client is None
            Exception: raise other exception
        """
        if self.influxdb_client is None:
            raise AttributeError('Please init_influxdb function run!!')

        try:
            query_api = self.influxdb_client.query_api()
            # query = f'import "influxdata/influxdb/schema" schema.fields(bucket: "{bucket}", measurement: "{measurement}")'
            query = f'import "influxdata/influxdb/schema"schema.measurementFieldKeys(bucket: "{bucket}",measurement: "{measurement}")'
            result = query_api.query(query)

            field_list = []
            for table in result:
                for record in table.records:
                    field_list.append(record.get_value())

            print(f'Bucket : {bucket},\nMeasurement : {measurement},\nField List : {field_list} \n')
        except Exception as e:
            raise Exception(e)

    def get_data_from_measurement(self, bucket, measurement, query_condition):
        """ Get data from influx measurement use query_condition into DataFrame.

        Args:
            bucket (str): influx bucket name
            measurement (str): influx measurement name
            query_condition (dict): query_condition

        Raises:
            AttributeError: raise when postgresdb client is None
            AttributeError: raise start_datetime in query_condition format is wrong.
            AttributeError: raise end_datetime in query_condition format is wrong.
            Exception: raise other exception

        Returns:
            pd.DataFrame: query result
        """        
        if self.influxdb_client is None:
            raise AttributeError('Please init_influxdb function run!!')

        if not isinstance(query_condition, Dict):
            raise AttributeError('query_condition parameter type is dictionary')

        start_datetime = query_condition.get('start_datetime')
        end_datetime = query_condition.get('end_datetime')

        limit_count = query_condition.get('limit_count')
        field_list = query_condition.get('column_list')

        # range query 조건
        range_query = '|> range(start: {}, stop: {})'
        if start_datetime and end_datetime:
            if not re.match(self.datetime_pattern, start_datetime):
                raise AttributeError('start_datetime pattern is wrong. pattern is "YYYY-MM-DD HH:mm:ss"')
            if not re.match(self.datetime_pattern, end_datetime):
                raise AttributeError('end_datetime pattern is wrong. pattern is "YYYY-MM-DD HH:mm:ss"')

        if start_datetime and end_datetime:
            range_query = range_query.format(datetime.strptime(start_datetime, '%Y-%m-%d %H:%M:%S').isoformat()+'Z', datetime.strptime(end_datetime, '%Y-%m-%d %H:%M:%S').isoformat()+'Z')
        else:
            range_query = range_query.format('-30d', 'now()')

        # field query 조건
        field_query = None
        if field_list:
            field_query = '|> filter(fn:(r) => {})'
            field_query_or_condition = 'r._field == "{}"'
            fin_field_query_or_condition = ''

            for i in range(0, len(field_list)):
                fin_field_query_or_condition += field_query_or_condition.format(field_list[i])

                if len(field_list) - (i+1) >= 1:
                    fin_field_query_or_condition = fin_field_query_or_condition + ' or '
            field_query = field_query.format(fin_field_query_or_condition)

        timezone_kst = timezone(timedelta(hours=9))
        query_api = self.influxdb_client.query_api()

        # bucket, range 조건, measurement, field 조건, limit 조건
        query_str = 'from(bucket:"{0}") {1} |> filter(fn:(r) => r._measurement == "{2}") {3} |> limit(n:{4}) |> sort(columns: ["_time"], desc: true)'.format(
            bucket, range_query, measurement, field_query if field_query else '', limit_count if limit_count else 1000)

        result = query_api.query(org=self.influx_organization, query=query_str)

        results = []
        for table in result:
            for record in table.records:
                results.append((record.get_field(), record.get_value(), record.get_time().astimezone(timezone_kst)))
        
        df = pd.DataFrame(results, columns=['field', 'value', 'time'])

        print('Current project name: {} \n Current device name : {}\n'.format(self.current_project_name, self.current_device))

        print('Data Preview \n')
        display(df.head())
        return df

    ###################################################### InfluxDB ######################################################

    def exception_handler(self, responseData, subtype):
        resp_dict = json.loads(responseData.content)
        if subtype == "500":
            errFormat = {
                "timestamp": resp_dict.get('timestamp'),
                "code": responseData.status_code,
                "error": resp_dict.get('error'),
                "message": resp_dict.get('error')
            }
        else:
            errFormat = {
                "timestamp": resp_dict.get('timestamp'),
                "code": resp_dict.get('code'),
                "error": resp_dict.get('error'),
                "message": resp_dict.get('error')
            }
        
        raise Exception(f"Failed!!!\n {errFormat}")

    def check_status_code(self, status_code):
        """ Check status code and return 0 or 1. 
            CheckStatusCode.FAILED.value(0) is fail.
            CheckStatusCode.OK.value(1) is 200(OK).
            HTTPStatusCode.CREATED.value(2) is 201(Created).
            CheckStatusCode.NO_CONTENT.value(3) is 204(No Content).

        Args:
            data (Dict): Response of api
            status_code (Int): Status code of resource
        """
        if status_code == HTTPStatusCode.INTERNAL_SERVER_ERROR.value:
            return CheckStatusCode.FAILED.value, f"Internal Server Error!!!, Status: {status_code}"
        elif status_code == HTTPStatusCode.OK.value:
            return CheckStatusCode.OK.value, f"Ok!!!, Status: {status_code}"
        elif status_code == HTTPStatusCode.CREATED.value:
            return CheckStatusCode.CREATED.value, f"Created!!!, Status: {status_code}"
        elif status_code == HTTPStatusCode.NO_CONTENT.value:
            return CheckStatusCode.NO_CONTENT.value, f"No Content!!!, Status: {status_code}"
        else:
            return CheckStatusCode.FAILED.value, ""

    def git_config_init(self):
        """ Gitea config Init
        """
        response = requests.request('GET', f"{self.giteamanager_url}/stackbase/v1/gitea-manager/users/{self.gitea_user_name}")

        if response.status_code == HTTPStatusCode.OK.value:
            print(f'set git config username: {self.gitea_user_name}, email: {self.email}')
            os.system(f'git config --global user.name {self.gitea_user_name}')
            os.system(f'git config --global user.email {self.email}')
            os.system(f'git config --global credential.helper store')
            os.system('git config --global init.defaultBranch main')
        else:
            os.system(f'git config --global --unset user.name')
            os.system(f'git config --global --unset user.email')
            os.system('git config --global init.defaultBranch main')
            print('git config user.name and user.email are replaced with user from the OS system.')

    ########################################################## s3 ##########################################################
    def s3_file_upload(self, upload_file_list):
        """ s3 file upload

        Args:
            local_path (str): current path(in Sagemaker notebook)
            bucket_name (str): bucket name for upload

        Raises:
            AttributeError: raise when s3 client is None
            clientError: raise when s3 upload_fileobject error exception
            e: raise when other exception
            FileNotFoundError: raise when local path file does not exist
        """
        try:
            pjm_timestamp = datetime.utcnow().strftime('%Y%m%d%H%M%S')  # timestamp(utc)
            message = f'{self.organizationId}{pjm_timestamp}'  # message(organizationId + timestamp)

            # hmac 생성(sha256)
            m = hmac.new(self.pjm_hmac_secret_key.encode('utf-8'), message.encode('utf-8'), hashlib.sha256)
            digest = m.digest()

            # header
            headers = {
                "Organization-Id": self.organizationId,
                "Timestamp": pjm_timestamp,
                "HMAC": base64.b64encode(digest).decode('utf-8')
            }

            data = {
                "s3_url": self.s3_url,
                "access_key": self.s3_access_key,
                "secret_key": self.s3_secret_key,
                "bucket_name": self.s3_bucket
            }

            files = []
            for f in upload_file_list:
                files.append(("files", open(f, "rb")))

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/s3/file-upload', headers=headers, data=data, files=files)

            if res.status_code == HTTPStatusCode.OK.value:
                print(res.status_code)
                return {"message": f"success file upload. {files}"}
            
            print(res.status_code, res.json().get('message'))
        except Exception as e:
            raise Exception(e)

    def s3_file_download(self, object_key, local_path):
        """ s3 file download

        Args:
            bucket_name (str): bucket name for download
            object_key (str): object key name for download
            local_path (str): current file path for download

        Raises:
            AttributeError: raise when s3 client is None
            clientError: raise when s3 download_fileobj exception
            e: raise when other exception
        """
        try:
            headers = self._generate_header()
            
            data = {
                "s3_url": self.s3_url,
                "access_key": self.s3_access_key,
                "secret_key": self.s3_secret_key,
                "bucket_name": self.s3_bucket,
                "download_file_name": object_key
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/s3/file-download', headers=headers, data=json.dumps(data))

            download_file_name = object_key.split('/')[1] if "/" in object_key else object_key

            if res.status_code == HTTPStatusCode.OK.value:
                with open(f'{local_path}/{download_file_name}', 'wb') as f:
                    f.write(res.content)
                return {"message": f"file download in {local_path}/{download_file_name}"}

            print(res.status_code, res.json().get('message'))
        except Exception as e:
            raise Exception(e)

    def s3_list_objet(self):
        """ s3 get object list of bucket

        Args:
            bucket_name (str): bucket name

        Raises:
            AttributeError: raise when s3 client is None
            clientError: raise when s3 client exception
            noCredentialsError: raise when no credential error
            e: raise when other exception

        Returns:
            _type_: _description_
        """
        try:
            headers = self._generate_header()

            data = {
                "s3_url": self.s3_url,
                "access_key": self.s3_access_key,
                "secret_key": self.s3_secret_key,
                "bucket_name": self.s3_bucket
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/s3/buckets/object', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data, columns=['file_name', 'size(Bytes)', 'time'])
                return df

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)

    def s3_delete_object(self, bucket_name, file_name):
        """ delete object of s3 bucket 

        Args:
            bucket_name (str): bucket name
            file_name (str): file name for delete

        Raises:
            AttributeError: raise when s3 client is None
            FileExistsError: raise when file does not exist in s3 bucket
            clientError: raise when s3 client exception
            noCredentialsError: raise when credential exception
            e: raise when other exception
        """
        try:
            headers = self._generate_header()

            data = {
                "s3_url": self.s3_url,
                "access_key": self.s3_access_key,
                "secret_key": self.s3_secret_key,
                "bucket_name": self.s3_bucket,
                "remove_file_name": file_name
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/s3/buckets/object/delete', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.NO_CONTENT.value:
                return res.status_code

            print(res.status_code, res.json().get('message'))
            return res.status_code
        except Exception as e:
            raise Exception(e)

    def s3_bucket_list(self):
        """ get s3 bucket list

        Raises:
            AttributeError: raise when s3 client is None
            clientError: raise when s3 client exception
            e: raise when other exception

        Returns:
            pd.DataFrame: query result
        """
        try:
            headers = self._generate_header()

            data = {
                "s3_url": self.s3_url,
                "access_key": self.s3_access_key,
                "secret_key": self.s3_secret_key
            }

            res = requests.request('POST', f'{self.endpointexplorer_url}/endpoint-explorer/v1/s3/buckets', headers=headers, data=json.dumps(data))

            if res.status_code == HTTPStatusCode.OK.value:
                result_data = res.json()
                df = pd.DataFrame(result_data, columns=['bucket_name', 'created_at'])
                return df

            print(res.status_code, res.json().get('message'))
            return None
        except Exception as e:
            raise Exception(e)
    ########################################################## s3 ##########################################################

    # 유저의 프로젝트 리스트 조회
    def get_projects(self):
        """ Print list of project in sdt cloud

        Raises:
            Exception: raise when other exception

        Returns:
            pd.DataFrame : query result
        """
        # PDTCLOUD-498 : ProjectManager hmac header 추가
        pjm_timestamp = datetime.utcnow().strftime('%Y%m%d%H%M%S')  # timestamp(utc)
        message = f'{self.organizationId}{pjm_timestamp}'  # message(organizationId + timestamp)

        # hmac 생성(sha256)
        m = hmac.new(self.pjm_hmac_secret_key.encode('utf-8'), message.encode('utf-8'), hashlib.sha256)
        digest = m.digest()

        # header
        headers = {
            "Content-Type": "application/json",
            "Organization-Id": self.organizationId,
            "Timestamp": pjm_timestamp,
            "HMAC": base64.b64encode(digest).decode('utf-8')
        }

        response = requests.request('get',f"{self.url}/pmt/projects", headers=headers)

        if response.status_code == HTTPStatusCode.OK.value:
            if len(response.content) > 0:
                result = json.loads(response.content)
                self.project_dataframe = pd.DataFrame(result.get('content'))
                return self.project_dataframe
            else:
                return None
        elif response.status_code == HTTPStatusCode.UNAUTHORIZED.value:
            ## refresh token
            pass


    def get_project_code_by_idx(self, idx, df):
        """
            Search project code with index in project_dataframe
        Args:
            idx (int): dataframe idx
            df (pd.DataFrame): project dataframe

        Returns:
            str: projectCode
        """        
        if df is not None and df.empty != True:
            return df.iloc[idx]['code'], df.iloc[idx]['name']

    # 프로젝트 선택
    def set_project(self, idx):
        """
            Check the connection information belonging to the project by looking up the projectCode using idx.
        Args:
            idx (int): project dataframe index

        """

        # PDTCLOUD-498 : ProjectManager hmac header 추가
        pjm_timestamp = datetime.utcnow().strftime('%Y%m%d%H%M%S')  # timestamp(utc)
        message = f'{self.organizationId}{pjm_timestamp}'  # message(organizationId + timestamp)

        # hmac 생성(sha256)
        m = hmac.new(self.pjm_hmac_secret_key.encode('utf-8'), message.encode('utf-8'), hashlib.sha256)
        digest = m.digest()

        # header
        headers = {
            "Content-Type": "application/json",
            "Organization-Id": self.organizationId,
            "Timestamp": pjm_timestamp,
            "HMAC": base64.b64encode(digest).decode('utf-8')
        }

        # project_dataframe에서 idx로 projectCode 가져오기
        projectCode, projectName = self.get_project_code_by_idx(idx, self.project_dataframe)

        response = requests.request('get',f"{self.url}/pmt/projects/{projectCode}/storage", headers=headers)
        respStatus, returnMessage = self.check_status_code(response.status_code)

        if respStatus == CheckStatusCode.FAILED.value:
            self.exception_handler(response, returnMessage)

        response_data = response.json()

        # minio 접속정보
        if response_data.get('bucket'):
            # self.s3_bucket = response_data.get('bucket').get('location')
            self.s3_access_key = response_data.get('bucket').get('minioAccessKey')
            self.s3_secret_key = response_data.get('bucket').get('minioSecretKey')
            self.s3_region = response_data.get('bucket').get('region')
            self.s3_bucket = response_data.get('bucket').get('name')

            bucket_df = pd.DataFrame.from_dict(data=response_data.get('bucket'), orient='index')

        database_data = response_data.get('database')

        for database in database_data:    
            if database.get('dbType') == 'TIMESTREAM':
                self.timestreamdb_access_key = database.get('accessKey')
                self.timestreamdb_secret_access_key = database.get('secretAccessKey')
                self.timestreamdb_region = database.get('region')
                self.timestreamdb_database = database.get('databaseName')
            elif database.get('dbType') == 'RDS_POSTGRES':
                self.postgresdb_url = database.get('url')
                self.postgresdb_database = database.get('databaseName')
                self.postgresdb_password = database.get('password')
                self.postgresdb_username = database.get('username')

        database_df = pd.DataFrame(data=response_data.get('database'))

        print('S3 Bucket Preview \n')
        display(bucket_df)
        print('\n\n')
        
        print('Database Preview \n')
        display(database_df)

    def get_user_repository(self):
        """ get user repository of gitea

        Returns:
            pd.DataFrame: query result
        """
        headers = {
            'email': self.email
        }

        response = requests.request('GET',f"{self.giteamanager_url}/stackbase/v1/gitea-manager/repos", headers=headers)

        if response.status_code == HTTPStatusCode.OK.value:
            repository_list = response.json().get('content')

            results = []
            if len(repository_list) > 0:
                for repo in repository_list:
                    results.append((repo.get('name'), repo.get('default_branch'), repo.get('clone_url')))

                df = pd.DataFrame(results, columns=['repository_name', 'branch', 'clone_url'])
                pd.set_option('display.max_colwidth', None)
                # df = df.style.format({'clone_url': self._make_clickable_url})
                html = df.to_html(render_links=True, escape=False)
                display(HTML(html))
            else:
                print('User Repository Does Not Exists')
        else:
            return f'Failed!, Status: {response.status_code}, Message: {response.json()}'

    def get_public_repository(self):
        """ get public repositry of gitea

        Returns:
            pd.DataFrame: query result
        """
        headers = {
            'Referer': 'sagemaker.aws'
        }
        response = requests.request('GET', f'{self.giteamanager_url}/stackbase/v1/gitea-manager/repos/all')
        
        if response.status_code == HTTPStatusCode.OK.value:
            repository_list = response.json().get('content')
            
            results = []
            if len(repository_list) > 0:
                for repo in repository_list:
                    results.append((repo.get('name'), repo.get('owner').get('email'), repo.get('default_branch'), repo.get('clone_url')))
                    
                df = pd.DataFrame(results, columns=['repository_name', 'owner_email', 'branch', 'clone_url'])
                pd.set_option('display.max_colwidth', None)
                # df = df.style.format({'clone_url': self._make_clickable_url})
                # df = df.style.format({'clone_url': self._make_clickable_url}, escape='html')
                html = df.to_html(render_links=True, escape=False)
                display(HTML(html))
            else:
                print('Repository List Does Not Exists')
        else:
            return f'Failed!, Status: {response.status_code}, Message: {response.json()}'

    def check_gitea_url_regex(self, url):
        pattern = re.compile(self.giteamanager_url_regex, re.IGNORECASE)
        result = pattern.match(url)
        
        if result:
            print('repository url verify!')
            return result.lastindex
        else:
            print('respotiroy url is not verify!')
            return 0

    def git_repo_init_folder(self, path='.'):
        """ git repository init in current path

        Args:
            path (str, optional): init .git folder . Defaults to '.'.

        Raises:
            AttributeError: raise when path is not directory
            FileNotFoundError: raise when path does not exist
            Exception: raise when other exception

        Returns:
            int: success status (200)
        """
        try:
            if path == '.':
                result = os.system('git init')
            else:
                if os.path.exists(path):
                    if os.path.isdir(path):
                        os.chdir(path)
                        result = os.system('git init')
                    else:
                        raise AttributeError(f'{path} is not directory!')
                else:
                    raise FileNotFoundError(f'{path} does not Exists')

            if result == 0:
                print('git Init Complete')
                return HTTPStatusCode.OK.value
        except Exception as e:
            raise Exception(e)

    def git_repo_add_remote_url(self, remote_url, path='.'):
        """ add repo remote gitea url

        Args:
            remote_url (str): gitea repository url
            path (str, optional): current path. Defaults to '.'.

        Raises:
            AttributeError: raise when path is not directory
            FileNotFoundError: raise when path does not exist
            Exception: raise when other exception

        Returns:
            int: status value(200, 204)
        """
        try:
            remote_ref = os.popen('git rev-parse --abbrev-ref --symbolic-full-name @{u} 2>/dev/null').read()

            if remote_ref == '':
                if path == '.':
                    result = os.system(f'git remote add --track main origin {remote_url}')
                else:
                    if os.path.exists(path):
                        if os.path.isdir(path):
                            os.chdir(path)
                            result = os.system(f'git remote add --track main origin {remote_url}')
                        else:
                            raise AttributeError(f'{path} is not directory!')
                    else:
                        raise FileNotFoundError(f'{path} does not Exists')

                if result == 0:
                    print('git remote_url add success!')
                    return HTTPStatusCode.OK.value
            else:
                print('remote_url already exists!')
                return HTTPStatusCode.NO_CONTENT.value
        except Exception as e:
            raise Exception(e)

    def clone_user_repository(self, clone_url):
        """ clone user gitea repository

        Args:
            clone_url (str): gitea repository url for clone

        Returns:
            int: clone status value (200, 500)
        """
        try:
            if clone_url != '':
                regex_result = self.check_gitea_url_regex(clone_url)
                if regex_result > 0:
                    result = os.system(f'git clone {clone_url}')

                    if result == 0:
                        repository_name = clone_url.split('/')[len(clone_url.split('/')) - 1].split('.')[0]

                        os.chdir(f'./{repository_name}')  # change directory
                        print(f'Completed Clone Repository!, Status: {HTTPStatusCode.OK.value}')
                        return HTTPStatusCode.OK.value
                    else:
                        print(f'Failed Clone Repository!, Status: {HTTPStatusCode.INTERNAL_SERVER_ERROR.value}')
                        return HTTPStatusCode.INTERNAL_SERVER_ERROR.value
        except Exception as e:
            print(f'Error Raise! Reason: {e}, Status: {HTTPStatusCode.INTERNAL_SERVER_ERROR.value}')
            return HTTPStatusCode.INTERNAL_SERVER_ERROR.value

    def add_stage_file(self, file_path_for_add_stages='.'):
        """ add stage file for commit gitea

        Args:
            file_path_for_add_stages (str, optional): file path for add stage. Defaults to '.'.

        Returns:
            int: status value(200, 500)
        """
        try:
            file_list = ' '.join(file_path_for_add_stages.split(','))
            result = os.system(f'git add {file_list}')

            if result == 0:
                print('Completed! Add File to Stage')
                return HTTPStatusCode.OK.value
            else:
                print('Failed! Add File to Stage')
                return HTTPStatusCode.INTERNAL_SERVER_ERROR.value
        except Exception as e:
            print(f'Error Raise! Reason: {e}, Status: {HTTPStatusCode.INTERNAL_SERVER_ERROR.value}')
            return HTTPStatusCode.INTERNAL_SERVER_ERROR.value

    def restore_staged_file(self, file_path_for_restore_stages):
        """ restore staged file

        Args:
            file_path_for_restore_stages (str): current path for restore staged

        Returns:
            int: value status(200, 500)
        """
        try:
            file_list = ' '.join(file_path_for_restore_stages.split(','))
            result = os.system(f'git restore --staged {file_list}')

            if result == 0:
                print('Completed! Restore File')
                return HTTPStatusCode.OK.value
            else:
                print('Failed! Restore File')
                return HTTPStatusCode.INTERNAL_SERVER_ERROR.value
        except Exception as e:
            print(f'Error Raise! Reason: {e}, Status: {HTTPStatusCode.INTERNAL_SERVER_ERROR.value}')
            return HTTPStatusCode.INTERNAL_SERVER_ERROR.value

    def commit_repository(self, commit_message):
        """ commit repository

        Args:
            commit_message (str): commit message

        Returns:
            int: value status(200, 500)
        """
        try:
            result = os.system(f'git commit -m "{commit_message}"')

            if result == 0:
                print('Commit Completed!')
                return HTTPStatusCode.OK.value
            else:
                print('Commit Failed!')
                return HTTPStatusCode.INTERNAL_SERVER_ERROR.value
        except Exception as e:
            print(f'Error Raise! Reason: {e}, Status: {HTTPStatusCode.INTERNAL_SERVER_ERROR.value}')
            return HTTPStatusCode.INTERNAL_SERVER_ERROR.value

    def push_repository(self, push_file_path='.'):
        """ push gitea repository 

        Args:
            push_file_path (str, optional): push file path for commit. Defaults to '.'.

        Raises:
            AttributeError: raise when current path remote url is None
            AttributeError: raise when gitea access_token does not exist
            Exception: raise when other exception

        Returns:
            int: status value(200, 500)
        """
        try:
            headers = {'email': self.email}
            response = requests.get(f"{self.giteamanager_url}/stackbase/v1/gitea-manager/users/me", headers=headers)
            
            if response.status_code == HTTPStatusCode.OK.value:
                access_token = response.json().get('accessToken')

                remote_push_url = os.popen('git config --get remote.origin.url').read().replace('\n', '')

                if self.check_gitea_url_regex(remote_push_url) > 0:
                    result = None
                    remote_push_url = remote_push_url.replace('http://', '')

                    # http://{username}:{access_token}@{gitea ip}/{username}/{repository_name}.git (self.minio_bucket : giteamanager 내 username과 동일)

                    if push_file_path == '.':
                        result = os.system(f'git push http://{self.minio_bucket}:{access_token}@{remote_push_url}')
                    else:
                        os.chdir(f'./{push_file_path}')
                        result = os.system(f'git push http://{self.minio_bucket}:{access_token}@{remote_push_url}')

                    if result == 0:
                        print('Push Completed!')
                        return HTTPStatusCode.OK.value
                    else:
                        print('Push Failed!')
                        return HTTPStatusCode.INTERNAL_SERVER_ERROR.value
                else:
                    raise AttributeError('Not added remote_url this path!')
            else:
                raise AttributeError('gitea access_token not exists!')
        except Exception as e:
            raise Exception(e)

    def pull_repository(self, repository_url):
        """ pull gitea repository

        Args:
            repository_url (str): gitea repository url

        Returns:
            int: status value(200, 500)
        """
        try:
            regex_result = self.check_gitea_url_regex(repository_url)

            if regex_result > 0:
                repository_name = repository_url.split('/')[len(repository_url.split('/')) - 1].split('.')[0]
                if os.path.exists(f'./{repository_name}'):
                    os.chdir(f'./{repository_name}')  # repository 폴더로 이동

                    result = os.system(f'git pull {repository_url} --allow-unrelated-histories')
                    if result == 0:
                        print(f'Completed! Pull Repository: {repository_url}')
                        return HTTPStatusCode.OK.value
                    else:
                        os.chdir('..')  # 상위 디렉토리로 이동
                        print(f'Failed! Pull Repository: {repository_url}')
                        return HTTPStatusCode.INTERNAL_SERVER_ERROR.value
                else:
                    print('repository path does not exist in notebook file path \n')
                    clone_repository_status = self.clone_user_repository(repository_url)
                    return clone_repository_status
        except Exception as e:
            print(f'Error Raise! Reason: {e}, Status: {HTTPStatusCode.INTERNAL_SERVER_ERROR.value}')
            return HTTPStatusCode.INTERNAL_SERVER_ERROR.value

    def create_repository(self, repo_name, private_flag=False):
        """ create gitea repository(public/private)

        Args:
            repo_name (str): repository name
            private_flag (bool, optional): private flag. Defaults to True.

        Raises:
            Exception: raise when conflict repository name
            Exception: raise when api exception
            Exception: raise when other exception

        Returns:
            str : returned clone url
        """
        try:
            headers = {'email': self.email}
            data = {
                "name": repo_name,
                "auto_init": False,
                "description": "",
                "gitignores": "",
                "issue_labels": "",
                "license": "",
                "private": True if private_flag else False,
                "readme": "Default",
                "template": False,
                "trust_model": "default"
            }

            response = requests.post(f"{self.giteamanager_url}/stackbase/v1/gitea-manager/repos", headers=headers, data=json.dumps(data))

            if response.status_code == HTTPStatusCode.CREATED.value:
                print('Repository Create Success! \n')
                response_data = response.json()
                return response_data.get('clone_url')
            elif response.status_code == HTTPStatusCode.CONFLICT.value:
                raise Exception(f'{repo_name} is conflict!')
            else:
                raise Exception(f'API Error Status : {response.status_code}')

        except Exception as e:
            raise Exception(e)
